import inspect

from predicthq.endpoints.base import UserBaseEndpoint
from predicthq.endpoints.decorators import accepts, returns
from .schemas import FeatureResultSet


class FeaturesEndpoint(UserBaseEndpoint):

    BASE_FEATURE_CRITERIA = {"stats": ["sum", "count"], "phq_rank": None}
    FIELDS_TO_MUTATE = frozenset(["phq_attendance_", "phq_viewership_sports", "phq_impact_severe_weather_"])

    @classmethod
    def mutate_bool_to_default_for_type(cls, user_request_spec):
        for key, val in user_request_spec.items():
            if any(key.startswith(x) for x in cls.FIELDS_TO_MUTATE) and isinstance(val, bool):
                user_request_spec[key] = cls.BASE_FEATURE_CRITERIA

    @accepts(query_string=False)
    @returns(FeatureResultSet)
    def obtain_features(self, **request):
        verify_ssl = request.pop("config", {}).get("verify_ssl", True)
        return self.client.post(
            self.build_url("v1", "features"),
            json=request,
            verify=verify_ssl,
        )



import functools
from collections import defaultdict

from pydantic import ValidationError as PydanticValidationError

from predicthq.endpoints.schemas import ResultSet
from predicthq.exceptions import ValidationError


def _to_url_params(data, glue=".", separator=","):
    """
    Converts data dictionary to url parameters
    """
    params = {}
    for key, value in data.items():
        if isinstance(value, bool):
            params[key] = 1 if value else 0
        elif isinstance(value, list):
            params[key] = separator.join(map(str, value))
        elif isinstance(value, dict):
            params.update(_flatten_dict(value, glue, separator, parent_key=key))
        else:
            params[key] = value
    return params


def _flatten_dict(d, glue, separator, parent_key=""):
    flat_dict = {}
    for k, v in d.items():
        if isinstance(v, dict):
            flat_dict.update(_flatten_dict(v, glue, separator, f"{parent_key}{glue}{k}" if parent_key else k))
            continue
        if isinstance(v, list):
            flat_dict.update({f"{parent_key}{glue}{k}" if parent_key else k: separator.join(map(str, v))})
            continue
        flat_dict.update({f"{parent_key}{glue}{k}" if parent_key else k: v})
    return flat_dict


def _process_kwargs(kwargs, separator="__"):
    data = defaultdict(dict)
    for key, value in kwargs.items():
        if separator in key:
            k, subk = key.split(separator)
            data[k][subk] = value
        else:
            data[key] = value
    return data



def accepts(query_string=True, role=None):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(endpoint, *args, **kwargs):
            data = _process_kwargs(kwargs)
            if hasattr(endpoint, "mutate_bool_to_default_for_type"):
                endpoint.mutate_bool_to_default_for_type(data)

            if query_string:
                data = _to_url_params(data=data)

            return f(endpoint, *args, **data)

        return wrapper

    return decorator


def returns(model_class):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(endpoint, *args, **kwargs):

            model = getattr(endpoint.Meta, f.__name__, {}).get("returns", model_class)

            data = f(endpoint, *args, **kwargs)
            try:
                loaded_model = model(**data)
                loaded_model._more = functools.partial(wrapper, endpoint)
                loaded_model._endpoint = endpoint
                return loaded_model
            except PydanticValidationError as e:
                raise ValidationError(e)

        return wrapper

    return decorator



class MetaEndpoint(type):
    def __new__(mcs, name, bases, data):
        if "Meta" not in data:

            class Meta:
                """ Used by decorators when overriding schema classes """

                pass

            data["Meta"] = Meta
        return super(MetaEndpoint, mcs).__new__(mcs, name, bases, data)


class BaseEndpoint(metaclass=MetaEndpoint):
    def __init__(self, client):
        self.client = client

    def build_url(self, prefix, suffix):
        return f"/{prefix.strip('/')}/{suffix.strip('/')}/"


class UserBaseEndpoint(BaseEndpoint):
    def __init__(self, client, account_id=None):
        self.account_id = account_id
        super(UserBaseEndpoint, self).__init__(client)

    def for_account(self, account_id):
        """
        Parameterised endpoint for account.

        Required when using a user access token, as the user may belong to multiple accounts with different plans

        """
        return self.__class__(self.client, account_id)

    def build_url(self, prefix, suffix):
        if self.account_id is not None:
            return f"/{prefix.strip('/')}/accounts/{self.account_id}/{suffix.strip('/')}/"
        else:
            return super(UserBaseEndpoint, self).build_url(prefix, suffix)



class PredictHQError(Exception):
    pass


class ConfigError(PredictHQError):
    pass


class APIError(PredictHQError):
    pass


class ClientError(APIError):
    pass


class ServerError(APIError):
    pass


class ValidationError(PredictHQError):
    pass



from predicthq.endpoints.base import BaseEndpoint
from predicthq.endpoints.decorators import accepts, returns
from .schemas import PlaceResultSet


class PlacesEndpoint(BaseEndpoint):
    @accepts()
    @returns(PlaceResultSet)
    def search(self, **params):
        verify_ssl = params.pop("config.verify_ssl", True)
        return self.client.get(
            self.build_url("v1", "places"),
            params=params,
            verify=verify_ssl,
        )



from pydantic import BaseModel


class LocationResult(BaseModel):
    lat: float
    lon: float


class SuggestedRadiusResultSet(BaseModel):
    radius: float
    radius_unit: str
    location: LocationResult



import os

import pytest

from predicthq.config import Config
from predicthq.exceptions import ConfigError


def test_defaults():
    config = Config()
    assert config.ENDPOINT_URL == "https://api.predicthq.com"
    assert config.OAUTH2_CLIENT_ID is None
    assert config.OAUTH2_CLIENT_SECRET is None
    assert config.OAUTH2_SCOPE is None
    assert config.OAUTH2_ACCESS_TOKEN is None
    assert config.LOGGING_LOG_LEVEL == "WARNING"


def test_defaults_from_locations():
    CONFIG_LOCATIONS = (
        os.path.join(os.path.dirname(__file__), "fixtures", "test.conf"),
        os.path.join(os.path.dirname(__file__), "fixtures", "test2.conf"),
    )

    config = Config(*CONFIG_LOCATIONS)
    assert config.ENDPOINT_URL == "https://example.org"
    assert config.OAUTH2_CLIENT_ID == "client_id"
    assert config.OAUTH2_CLIENT_SECRET == "client_secret"
    assert config.OAUTH2_SCOPE == "account events"
    assert config.OAUTH2_ACCESS_TOKEN == "access_token"
    assert config.LOGGING_LOG_LEVEL == "DEBUG"


def test_defaults_from_environment():
    os.environ["PREDICTHQ_ENDPOINT_URL"] = "https://example.org/endpoint/"
    try:
        config = Config(os.path.join(os.path.dirname(__file__), "fixtures", "test.conf"))
        assert config.ENDPOINT_URL == "https://example.org/endpoint/"
        assert config.OAUTH2_CLIENT_ID == "client_id"
        assert config.OAUTH2_CLIENT_SECRET == "client_secret"
        assert config.OAUTH2_SCOPE == "scope"
        assert config.OAUTH2_ACCESS_TOKEN == "access_token"
        assert config.LOGGING_LOG_LEVEL == "log_level"
    finally:
        del os.environ["PREDICTHQ_ENDPOINT_URL"]


def test_get_config():
    config = Config()
    assert config.ENDPOINT_URL == "https://api.predicthq.com"

    with pytest.raises(ConfigError):
        assert config.INVALID_CONFIG


def test_set_config():
    config = Config()
    config.ENDPOINT_URL = "https://example.org"
    assert config.ENDPOINT_URL == "https://example.org"

    with pytest.raises(ConfigError):
        config.INVALID_CONFIG = "invalid"



from datetime import datetime
from typing import Optional

from pydantic import BaseModel


class Industry(BaseModel):
    id: str
    name: str


class Account(BaseModel):
    id: str
    name: str
    description: Optional[str] = None
    industry: Optional[Industry] = None
    created_at: datetime
    updated_at: datetime



from typing import List, Optional, Union

from pydantic import BaseModel, field_validator

from predicthq.config import config


class AccessToken(BaseModel):
    access_token: str
    token_type: str
    scope: List[str]
    refresh_token: Optional[str] = None
    expires_in: Optional[int] = None

    @field_validator("scope", mode="before")
    def parse_scope(cls, value: Optional[Union[List[str], str]] = None):
        if isinstance(value, str):
            if " " in value:
                return value.split(" ")
            elif "," in value:
                return value.split(",")
            return [value]
        return value



import functools

from predicthq.config import config


DEFAULT_CONFIG_MAPPING = {
    "client_id": config.OAUTH2_CLIENT_ID,
    "client_secret": config.OAUTH2_CLIENT_SECRET,
    "scope": config.OAUTH2_SCOPE,
    "grant_type": "client_credentials",
    "token_type_hint": "access_token",
}


def preload_config_defaults(defaults_to_load):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(endpoint, *args, **kwargs):
            for key in defaults_to_load:
                if key not in kwargs and key in DEFAULT_CONFIG_MAPPING:
                    kwargs[key] = DEFAULT_CONFIG_MAPPING[key]

            if kwargs.get("scope") and "," in kwargs["scope"]:
                kwargs["scope"] = " ".join(kwargs["scope"].split(","))

            return f(endpoint, *args, **kwargs)

        return wrapper

    return decorator



import json
import logging
from urllib.parse import urljoin, urlparse, urlunparse
from weakref import proxy

import requests

from .config import config
from .exceptions import ClientError, ServerError
from .version import __version__


class Client(object):
    @classmethod
    def build_url(cls, path):
        result = list(urlparse(path))
        result[2] = f"/{result[2].strip('/')}/"
        return urljoin(config.ENDPOINT_URL, urlunparse(result))

    def __init__(self, access_token=None):
        self.logger = logging.getLogger("predicthq")
        self.logger.setLevel(config.LOGGING_LOG_LEVEL)
        self.access_token = access_token or config.OAUTH2_ACCESS_TOKEN
        self.initialize_endpoints()

    def initialize_endpoints(self):
        from predicthq import endpoints

        self.oauth2 = endpoints.OAuth2Endpoint(proxy(self))
        self.broadcasts = endpoints.BroadcastsEndpoint(proxy(self))
        self.events = endpoints.EventsEndpoint(proxy(self))
        self.features = endpoints.FeaturesEndpoint(proxy(self))
        self.accounts = endpoints.AccountsEndpoint(proxy(self))
        self.places = endpoints.PlacesEndpoint(proxy(self))
        self.radius = endpoints.SuggestedRadiusEndpoint(proxy(self))

    def get_headers(self, headers):
        _headers = {"Accept": "application/json", "x-user-agent": f"PHQ-Py-SDK/{__version__}"}
        if self.access_token:
            _headers.update(
                {
                    "Authorization": f"Bearer {self.access_token}",
                }
            )
        _headers.update(**headers)
        return _headers

    def request(self, method, path, **kwargs):
        headers = self.get_headers(kwargs.pop("headers", {}))
        response = requests.request(method, self.build_url(path), headers=headers, **kwargs)
        self.logger.debug(response.request.url)
        try:
            response.raise_for_status()
        except requests.HTTPError:
            try:
                error = response.json()
            except ValueError:
                error = response.content

            if 400 <= response.status_code <= 499:
                raise ClientError(error)
            else:
                raise ServerError(error)

        try:
            result = response.json() or None
        except:
            result = None

        if self.logger.getEffectiveLevel() == logging.DEBUG:
            self.logger.debug(json.dumps(result, indent=True))

        return result

    def get(self, path, **kwargs):
        return self.request("get", path, **kwargs)

    def post(self, path, **kwargs):
        return self.request("post", path, **kwargs)

    def put(self, path, **kwargs):
        return self.request("put", path, **kwargs)

    def patch(self, path, **kwargs):
        return self.request("patch", path, **kwargs)

    def delete(self, path, **kwargs):
        return self.request("delete", path, **kwargs)

    def authenticate(self, **kwargs):
        token = self.oauth2.get_token(**kwargs)
        self.access_token = token.access_token
        return token



from predicthq.endpoints.base import UserBaseEndpoint
from predicthq.endpoints.decorators import accepts, returns
from .schemas import (
    EventResultSet,
    CountResultSet,
    CalendarResultSet,
)


class EventsEndpoint(UserBaseEndpoint):
    @accepts()
    @returns(EventResultSet)
    def search(self, **params):
        verify_ssl = params.pop("config.verify_ssl", True)
        return self.client.get(
            self.build_url("v1", "events"),
            params=params,
            verify=verify_ssl,
        )

    @accepts()
    @returns(CountResultSet)
    def count(self, **params):
        verify_ssl = params.pop("config.verify_ssl", True)
        return self.client.get(
            self.build_url("v1", "events/count"),
            params=params,
            verify=verify_ssl,
        )

    @accepts()
    @returns(CalendarResultSet)
    def calendar(self, **params):
        verify_ssl = params.pop("config.verify_ssl", True)
        return self.client.get(
            self.build_url("v1", "events/calendar"),
            params=params,
            verify=verify_ssl,
        )



from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel

from predicthq.endpoints.schemas import ResultSet


class GeoPoint(BaseModel):
    lat: float
    lon: float


class BroadcastEventEntities(BaseModel):
    entity_id: str
    type: str
    name: str
    formatted_address: Optional[str] = None


class BroadcastEventLocation(BaseModel):
    geopoint: GeoPoint
    place_hierarchies: List[List[str]]
    country: str


class BroadcastEventDates(BaseModel):
    start: datetime
    end: Optional[datetime] = None
    start_local: datetime
    end_local: Optional[datetime] = None
    timezone: str

    # predicted_end_local is a paid feature.
    # It will only show up in your response body if you
    # have subscribed to it.
    predicted_end_local: Optional[datetime] = None


class BroadcastEvent(BaseModel):
    event_id: str
    title: str
    category: str
    labels: List[str]
    dates: BroadcastEventDates
    location: BroadcastEventLocation
    entities: List[BroadcastEventEntities]

    # The following fields are paid features.
    # They will only show up in your response body if you
    # have subscribed to them.
    phq_attendance: Optional[int] = None
    phq_rank: Optional[int] = None
    local_rank: Optional[int] = None
    aviation_rank: Optional[int] = None


class Place(BaseModel):
    place_id: str
    type: str
    name: str
    county: str
    region: str
    country: str


class BroadcastLocation(BaseModel):
    geopoint: GeoPoint
    place_hierarchies: List[List[str]]
    places: List[Place]
    country: str


class BroadcastDates(BaseModel):
    start: datetime
    start_local: datetime
    timezone: str


class Broadcast(BaseModel):
    broadcast_id: str
    updated: datetime
    first_seen: datetime
    dates: Optional[BroadcastDates] = None
    location: Optional[BroadcastLocation] = None
    phq_viewership: Optional[int] = None
    record_status: str
    broadcast_status: Optional[str] = None
    event: Optional[BroadcastEvent] = None


class BroadcastResultSet(ResultSet):
    overflow: bool
    results: List[Optional[Broadcast]]



from predicthq.endpoints.base import BaseEndpoint
from predicthq.endpoints.decorators import accepts, returns
from .decorators import preload_config_defaults
from .schemas import AccessToken


class OAuth2Endpoint(BaseEndpoint):
    @accepts()
    @preload_config_defaults(["client_id", "client_secret", "scope", "grant_type"])
    @returns(AccessToken)
    def get_token(self, client_id, client_secret, scope, grant_type, **kwargs):
        verify_ssl = kwargs.pop("config.verify_ssl", True)
        data = {
            "grant_type": grant_type,
            "scope": scope,
        }
        data.update(kwargs)
        return self.client.post(
            "/oauth2/token/",
            auth=(client_id, client_secret),
            data=data,
            verify=verify_ssl,
        )

    @accepts()
    @preload_config_defaults(["client_id", "client_secret", "token_type_hint"])
    def revoke_token(self, client_id, client_secret, token, token_type_hint, **kwargs):
        verify_ssl = kwargs.pop("config.verify_ssl", True)
        data = {
            "token_type_hint": token_type_hint,
            "token": token,
        }
        self.client.post(
            "/oauth2/revoke/",
            auth=(client_id, client_secret),
            data=data,
            verify=verify_ssl,
        )



import pytest

import predicthq
from predicthq import endpoints
from predicthq.endpoints.oauth2.schemas import AccessToken
from predicthq.exceptions import ClientError, ServerError
from tests import with_mock_responses, with_config, with_mock_client, load_fixture


class ClientTest(object):
    def setUp(self):
        self.client = predicthq.Client()

    @with_config(ENDPOINT_URL="https://api.predicthq.com")
    def test_build_url(self):
        assert self.client.build_url("v1") == "https://api.predicthq.com/v1/"

    def test_endpoints_initialization(self):
        assert isinstance(self.client.oauth2, endpoints.OAuth2Endpoint)
        assert isinstance(self.client.accounts, endpoints.AccountsEndpoint)
        assert isinstance(self.client.events, endpoints.EventsEndpoint)
        assert isinstance(self.client.places, endpoints.PlacesEndpoint)

    @with_mock_responses()
    def test_request(self, responses):
        self.client.logger.setLevel("DEBUG")

        assert self.client.get("/v1/") == responses.calls[0].response.json()

        assert self.client.get("/no-content/") is None

        assert self.client.get("/invalid-json/") is None

        with pytest.raises(ClientError) as excinfo:
            self.client.get("/not-found/")
        assert excinfo.value.args[0] == responses.calls[3].response.json()

        with pytest.raises(ServerError) as excinfo:
            self.client.get("/server-error/")
        assert excinfo.value.args[0] == responses.calls[4].response.json()

        with pytest.raises(ServerError) as excinfo:
            self.client.get("/no-json/")
        assert excinfo.value.args[0] == responses.calls[5].response.content

        # Test headers
        self.client.authenticate(client_id="client_id", client_secret="client_secret", scope=["account"])
        assert responses.calls[6].request.headers["Authorization"] == "Basic Y2xpZW50X2lkOmNsaWVudF9zZWNyZXQ="
        assert responses.calls[6].request.headers["Content-Type"] == "application/x-www-form-urlencoded"

        self.client.accounts.self()
        assert responses.calls[7].request.headers["Authorization"] == "Bearer token123"
        assert responses.calls[7].request.headers["Accept"] == "application/json"

    @with_mock_client(request_returns={"result": "value"})
    def test_get(self, client):
        result = self.client.get("/get/", params={"q": "query"})
        client.request.assert_called_once_with("get", "/get/", params={"q": "query"})
        assert result == client.request.return_value

    @with_mock_client(request_returns={"result": "value"})
    def test_post(self, client):
        result = self.client.post("/post/", data={"key": "value"})
        client.request.assert_called_once_with("post", "/post/", data={"key": "value"})
        assert result == client.request.return_value

    @with_mock_client(request_returns={"result": "value"})
    def test_put(self, client):
        result = self.client.put("/put/", data={"key": "value"})
        client.request.assert_called_once_with("put", "/put/", data={"key": "value"})
        assert result == client.request.return_value

    @with_mock_client(request_returns={"result": "value"})
    def test_patch(self, client):
        result = self.client.patch("/patch/", data={"key": "value"})
        client.request.assert_called_once_with("patch", "/patch/", data={"key": "value"})
        assert result == client.request.return_value

    @with_mock_client()
    def test_delete(self, client):
        result = client.delete("/delete/")
        client.request.assert_called_once_with("delete", "/delete/")
        assert result == client.request.return_value

    @with_mock_client(request_returns=load_fixture("access_token"))
    def test_authenticate(self, client):
        token = self.client.authenticate(
            client_id="client_id", client_secret="client_secret", scope=["account", "events"]
        )
        client.request.assert_called_once_with(
            "post",
            "/oauth2/token/",
            auth=("client_id", "client_secret"),
            data={"scope": "account events", "grant_type": "client_credentials"},
        )

        assert isinstance(token, AccessToken)
        assert token.to_primitive() == client.request.return_value

        assert self.client.access_token == "token123"

    def test_construct_with_access_token(self):
        client = predicthq.Client(access_token="token123")
        assert client.access_token == "token123"

    @with_config(OAUTH2_ACCESS_TOKEN="token123")
    def test_construct_with_access_token_config(self):
        client = predicthq.Client()
        assert client.access_token == "token123"



from datetime import date
from typing import List

from typing import Optional

from pydantic import BaseModel

from predicthq.endpoints.schemas import ResultSet


class FeatureRankLevel(BaseModel):
    rank_levels: dict


class FeatureStats(BaseModel):
    avg: Optional[float] = None
    count: Optional[int] = None
    max: Optional[float] = None
    median: Optional[float] = None
    min: Optional[float] = None
    sum: Optional[float] = None
    std_dev: Optional[float] = None


class FeatureStat(BaseModel):
    stats: FeatureStats


class Feature(BaseModel):
    class Options:
        serialize_when_none = False

    date: date
    # Attendance based features
    phq_attendance_academic_graduation: Optional[FeatureStat] = None
    phq_attendance_academic_social: Optional[FeatureStat] = None
    phq_attendance_community: Optional[FeatureStat] = None
    phq_attendance_concerts: Optional[FeatureStat] = None
    phq_attendance_conferences: Optional[FeatureStat] = None
    phq_attendance_expos: Optional[FeatureStat] = None
    phq_attendance_festivals: Optional[FeatureStat] = None
    phq_attendance_performing_arts: Optional[FeatureStat] = None
    phq_attendance_school_holidays: Optional[FeatureStat] = None
    phq_attendance_sports: Optional[FeatureStat] = None
    # Rank based features
    phq_rank_daylight_savings: Optional[FeatureRankLevel] = None
    phq_rank_health_warnings: Optional[FeatureRankLevel] = None
    phq_rank_observances: Optional[FeatureRankLevel] = None
    phq_rank_public_holidays: Optional[FeatureRankLevel] = None
    phq_rank_school_holidays: Optional[FeatureRankLevel] = None
    phq_rank_politics: Optional[FeatureRankLevel] = None
    phq_rank_academic_session: Optional[FeatureRankLevel] = None
    phq_rank_academic_exam: Optional[FeatureRankLevel] = None
    phq_rank_academic_holiday: Optional[FeatureRankLevel] = None
    # Impact based feature criteria
    phq_impact_severe_weather_air_quality_retail: Optional[FeatureStat] = None
    phq_impact_severe_weather_blizzard_retail: Optional[FeatureStat] = None
    phq_impact_severe_weather_cold_wave_retail: Optional[FeatureStat] = None
    phq_impact_severe_weather_cold_wave_snow_retail: Optional[FeatureStat] = None
    phq_impact_severe_weather_cold_wave_storm_retail: Optional[FeatureStat] = None
    phq_impact_severe_weather_dust_retail: Optional[FeatureStat] = None
    phq_impact_severe_weather_dust_storm_retail: Optional[FeatureStat] = None
    phq_impact_severe_weather_flood_retail: Optional[FeatureStat] = None
    phq_impact_severe_weather_heat_wave_retail: Optional[FeatureStat] = None
    phq_impact_severe_weather_hurricane_retail: Optional[FeatureStat] = None
    phq_impact_severe_weather_thunderstorm_retail: Optional[FeatureStat] = None
    phq_impact_severe_weather_tornado_retail: Optional[FeatureStat] = None
    phq_impact_severe_weather_tropical_storm_retail: Optional[FeatureStat] = None
    # Viewership based features
    phq_viewership_sports: Optional[FeatureStat] = None
    phq_viewership_sports_american_football: Optional[FeatureStat] = None
    phq_viewership_sports_american_football_ncaa_men: Optional[FeatureStat] = None
    phq_viewership_sports_american_football_nfl: Optional[FeatureStat] = None
    phq_viewership_sports_auto_racing: Optional[FeatureStat] = None
    phq_viewership_sports_auto_racing_indy_car: Optional[FeatureStat] = None
    phq_viewership_sports_auto_racing_nascar: Optional[FeatureStat] = None
    phq_viewership_sports_baseball: Optional[FeatureStat] = None
    phq_viewership_sports_baseball_mlb: Optional[FeatureStat] = None
    phq_viewership_sports_baseball_ncaa_men: Optional[FeatureStat] = None
    phq_viewership_sports_basketball: Optional[FeatureStat] = None
    phq_viewership_sports_basketball_ncaa_women: Optional[FeatureStat] = None
    phq_viewership_sports_basketball_ncaa_men: Optional[FeatureStat] = None
    phq_viewership_sports_basketball_nba: Optional[FeatureStat] = None
    phq_viewership_sports_boxing: Optional[FeatureStat] = None
    phq_viewership_sports_golf: Optional[FeatureStat] = None
    phq_viewership_sports_golf_masters: Optional[FeatureStat] = None
    phq_viewership_sports_golf_pga_championship: Optional[FeatureStat] = None
    phq_viewership_sports_golf_pga_tour: Optional[FeatureStat] = None
    phq_viewership_sports_golf_us_open: Optional[FeatureStat] = None
    phq_viewership_sports_horse_racing: Optional[FeatureStat] = None
    phq_viewership_sports_horse_racing_belmont_stakes: Optional[FeatureStat] = None
    phq_viewership_sports_horse_racing_kentucky_derby: Optional[FeatureStat] = None
    phq_viewership_sports_horse_racing_preakness_stakes: Optional[FeatureStat] = None
    phq_viewership_sports_ice_hockey: Optional[FeatureStat] = None
    phq_viewership_sports_ice_hockey_nhl: Optional[FeatureStat] = None
    phq_viewership_sports_mma: Optional[FeatureStat] = None
    phq_viewership_sports_mma_ufc: Optional[FeatureStat] = None
    phq_viewership_sports_soccer: Optional[FeatureStat] = None
    phq_viewership_sports_soccer_concacaf_champions_league: Optional[FeatureStat] = None
    phq_viewership_sports_soccer_concacaf_gold_cup: Optional[FeatureStat] = None
    phq_viewership_sports_soccer_copa_america_men: Optional[FeatureStat] = None
    phq_viewership_sports_soccer_fifa_world_cup_women: Optional[FeatureStat] = None
    phq_viewership_sports_soccer_fifa_world_cup_men: Optional[FeatureStat] = None
    phq_viewership_sports_soccer_mls: Optional[FeatureStat] = None
    phq_viewership_sports_soccer_uefa_champions_league_men: Optional[FeatureStat] = None
    phq_viewership_sports_softball: Optional[FeatureStat] = None
    phq_viewership_sports_softball_ncaa_women: Optional[FeatureStat] = None
    phq_viewership_sports_tennis: Optional[FeatureStat] = None
    phq_viewership_sports_tennis_us_open: Optional[FeatureStat] = None
    phq_viewership_sports_tennis_wimbledon: Optional[FeatureStat] = None


class FeatureResultSet(ResultSet):
    results: List[Optional[Feature]]



from typing import List, Optional, Tuple

from pydantic import BaseModel, field_validator

from predicthq.endpoints.schemas import ResultSet


class Place(BaseModel):
    id: str
    type: str
    name: str
    county: Optional[str] = None
    region: Optional[str] = None
    country: Optional[str] = None
    country_alpha2: Optional[str] = None
    country_alpha3: Optional[str] = None
    location: Tuple[float, float]

    @field_validator("location")
    def reverse_location(cls, value: Tuple[float, float]):
        """
        Location is received as follows:
        "location": [lon, lat]
        By convention, we are used to have geopoints formated as lat,lon, which is what this validator enforces
        """
        return value[1], value[0]


class PlaceResultSet(ResultSet):
    results: List[Optional[Place]]



from .oauth2 import OAuth2Endpoint
from .v1.accounts import AccountsEndpoint
from .v1.broadcasts import BroadcastsEndpoint
from .v1.events import EventsEndpoint
from .v1.features import FeaturesEndpoint
from .v1.places import PlacesEndpoint
from .v1.radius import SuggestedRadiusEndpoint


__all__ = [
    "OAuth2Endpoint",
    "AccountsEndpoint",
    "BroadcastsEndpoint",
    "EventsEndpoint",
    "FeaturesEndpoint",
    "PlacesEndpoint",
    "SuggestedRadiusEndpoint",
]



from urllib.parse import parse_qsl, urlparse
from typing import Callable, Optional

from pydantic import BaseModel, HttpUrl


class _Config(BaseModel):
    verify_ssl: Optional[bool] = False


class ConfigMixin(BaseModel):
    config: _Config


class ResultSet(BaseModel):
    _endpoint: Optional[Callable] = None
    _more: Optional[Callable] = None

    count: Optional[int] = None
    previous: Optional[HttpUrl] = None
    next: Optional[HttpUrl] = None

    def _parse_params(self, url):
        return dict(url.query_params())

    def has_previous(self):
        return self.previous is not None

    def has_next(self):
        return self.next is not None

    def get_next(self):
        if not self.has_next() or not hasattr(self, "_more"):
            return
        params = self._parse_params(self.next)
        return self._more(**params)

    def get_previous(self):
        if not self.has_previous() or not hasattr(self, "_more"):
            return
        params = self._parse_params(self.previous)
        return self._more(**params)

    def iter_pages(self):
        next_page = self
        while next_page.has_next():
            next_page = next_page.get_next()
            yield next_page

    def iter_items(self):
        return iter(self.results)

    def iter_all(self):
        for item in self.iter_items():
            yield item
        for page in self.iter_pages():
            for item in page.iter_items():
                yield item

    def __iter__(self):
        return self.iter_items()



from datetime import date, datetime
from typing import List, Optional, Tuple, Union

from pydantic import BaseModel, field_validator

from predicthq.endpoints.schemas import ResultSet


class Entities(BaseModel):
    entity_id: str
    name: str
    type: str
    formatted_address: Optional[str] = None

class PredictedEventSpendIndustries(BaseModel):
    accommodation: int
    hospitality: int
    transportation: int


class Point(BaseModel):
    type: str
    coordinates: List[float]


class MultiPoint(BaseModel):
    type: str
    coordinates: List[List[float]]


class Polygon(BaseModel):
    type: str
    coordinates: List[List[List[float]]]


class MultiPolygon(BaseModel):
    type: str
    coordinates: List[List[List[List[float]]]]


class Geo(BaseModel):
    geometry: Union[Point, MultiPoint, Polygon, MultiPolygon]
    placekey: Optional[str] = None


class ParentEvent(BaseModel):
    parent_event_id: str


class ImpactPatternImpacts(BaseModel):
    date_local: date
    value: int
    position: str


class ImpactPattern(BaseModel):

    vertical: str
    impact_type: str
    impacts: List[ImpactPatternImpacts]


class Event(BaseModel):
    cancelled: Optional[datetime] = None
    category: str
    country: str
    deleted_reason: Optional[str] = None
    description: Optional[str] = None
    duplicate_of_id: Optional[str] = None
    duration: Optional[int] = None
    end: Optional[datetime] = None
    first_seen: Optional[datetime] = None
    geo: Optional[Geo] = None
    id: str
    impact_patterns: Optional[List[ImpactPattern]] = []
    labels: List[str]
    location: Optional[Tuple[float, float]] = None
    parent_event: Optional[ParentEvent] = None
    place_hierarchies: Optional[List[List[str]]] = None
    postponed: Optional[datetime] = None
    relevance: Optional[float] = None
    scope: Optional[str] = None
    start: datetime
    state: Optional[str] = None
    timezone: Optional[str] = None
    title: str
    updated: Optional[datetime] = None

    # The below fields are only available if they are enabled in your plan.
    aviation_rank: Optional[int] = None  # Aviation Rank add-on
    brand_safe: Optional[bool] = None
    entities: Optional[List[Entities]] = []  # Venues and addresses add-on
    local_rank: Optional[int] = None  # Local Rank add-on
    phq_attendance: Optional[int] = None  # PHQ Attendance add-on
    predicted_end: Optional[datetime] = None
    private: Optional[bool] = None  # Loop add-on
    rank: Optional[int] = None  # PHQ Rank add-on
    predicted_event_spend: Optional[int] = None  # Predicted Event Spend add-on
    predicted_event_spend_industries: Optional[PredictedEventSpendIndustries] = None  # Predicted Event Spend add-on

    @field_validator("location")
    def reverse_location(cls, value: Tuple[float, float]):
        """
        Location is received as follows:
        "location": [lon, lat]
        By convention, we are used to have geopoints formated as lat,lon, which is what this validator enforces
        """
        return value[1], value[0]


class EventResultSet(ResultSet):
    overflow: Optional[bool] = False
    results: List[Optional[Event]]


class CountResultSet(BaseModel):
    count: int
    top_rank: float
    rank_levels: dict
    categories: dict
    labels: dict


class CalendarDay(BaseModel):
    date: date
    count: int
    top_rank: float
    rank_levels: dict
    categories: dict
    labels: dict
    top_events: EventResultSet


class CalendarResultSet(ResultSet):
    results: List[Optional[CalendarDay]]



#!/usr/bin/env python

import os
import codecs
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    filename = os.path.join(os.path.dirname(__file__), *parts)
    with codecs.open(filename, encoding="utf-8") as fp:
        return fp.read()


about = {}

# Load version number dynamically
with open(os.path.join(here, 'predicthq', 'version.py'), 'r') as f:
    exec(f.read(), about)


REPO_URL = "https://github.com/predicthq/sdk-py"
PYPI_README_NOTE = f"For the latest source, discussions, bug reports, etc., please visit the [GitHub repository]({REPO_URL})"
LONG_DESCRIPTION = "\n\n".join([PYPI_README_NOTE, read("README.md")])


setup(
    name="predicthq",
    version=about['__version__'],
    description="PredictHQ Event Intelligence",
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    license="MIT",
    author="PredictHQ",
    author_email="developers@predicthq.com",
    url=REPO_URL,
    packages=find_packages(exclude=("tests*",)),
    python_requires=">=3.7",
    install_requires=[
        "pydantic>=2,<3",
        "requests>=2.7.0",
        "python-dateutil>=2.4.2",
        "pytz>=2017.2,<=2021.1",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Environment :: Web Environment",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]
)
